// Name: Jonathan Turner
// Class: CS 4306/01
// Term: Spring 2024
// Instructor: Dr. Haddad
// Assignment: 1
// IDE Name: IntelliJ
/*
Algorithm Design Block
    Algorithm title: Locker Doors Problem

    Logical steps:
        Step 1: Declare two fields, An Array of Assignments.Assignment1.Locker & A Comparisons Integer.
        Step 2: Gather user input for the number of lockers in the problem.
        Step 3: Start the locker swap. Check if the input has been gathered prior to execution, if not stop and revert to Step 2.
        Step 4: Set comparisons to 0.
        Step 5: Start the outer loop (i) that begins at 1 and ends at n, incremented by 1.
        Step 6: Start the inner loop (j) that begins at base and ends at n, incremented by the i. (j+=i)
        Step 7: Check lockers[j], if open close it, if closed open it.
        Step 8: Inside the j loop, increment comparisons by 1.
        Step 9: After running the locker swapping, check which lockers are open and closed.
        Step 10: Print out N, lockers that are open, lockers that are closed, and the comparisons.

    Pseudocode Syntax:
        size <- input
        locker <- new locker[input]
        comparisons <- 0
        for i <- 1 to n
            for j <- i to n by j+=i
                if locker[j] is False // Basic Operation
                    locker[j] <- True
                else
                    locker[i] <- False
                comparisons++
            end
        end

        open_lockers <- {}
        closed_lockers <- {}
        for i <- 1 to n
            if locker[i] is False
                closed_lockers.add(i)
            else
                open_lockers.add(i)
        end

        print("Number of Lockers:   " + size)
        print("Open lockers:        " + open_lockers)
        print("Closed lockers:      " + closed_lockers)
        print("Comparisons:         " + comparisons)
*/
package Assignments.A1;

import java.util.Scanner;

public class LockerDoors {

    /**
     * Starts the program and begins the sentinel loop.
     */
    public static void main(String[] args) {
        LockerDoors program = new LockerDoors();
        program.start();
    }

    // The Field Variables
    private Locker[] lockers;
    private int comparisons;

    /**
     * Controls the sentinal loop. Begins with getting an option, executing it, and repeating.
     * Only exits when option is 3.
     */
    public void start() {
        int option = getOption();
        while (option != 3) {
            executeOption(option);
            option = getOption();
        }
    }

    /**
     * Prints out the menu of options, asks for an input, and if that input is invalid it prints an error and
     * prompts the user again for the input.
     */
    private int getOption() {
        displayMenu(); // Prints out the menu

        // Asks for the option and begins the loop until a valid option is gathered.
        System.out.print("Enter option number: ");
        int input = getIntegerInput();
        while (input < 1 || input > 3) { // Compares it to the valid options available.
            System.out.println("\nPlease enter a valid input.");
            System.out.print("Enter option number: ");
            input = getIntegerInput();
        }
        System.out.println();
        return input;
    }

    /**
     * Gets user input to a number. A prompt must be provided prior to running this method.
     * If the input was not an integer/(unable to be autoboxed), returns -1.
     */
    private int getIntegerInput() {
        Scanner sc = new Scanner(System.in);
        try {
            String textInput = sc.nextLine();
            return Integer.parseInt(textInput);
        } catch (Exception e) {
            return -1;
        }
    }

    /**
     * This method executes the option based on that which was provided in the parameters.
     * This method prevents any option from running (Ex. Option 2) until all preconditions have been met.
     * Such as setting a size of lockers prior to running the algorithm on the lockers.
     */
    private void executeOption(int option) {
        if (option == 1) { // Sets the size of lockers based on user input.
            System.out.print("Enter the number of lockers: ");
            int input = getIntegerInput(); // Gets the user's input for the number of lockers

            // If the input is non-positive then it loops until a positive value is presented.
            while (input < 1) {
                System.out.println("\nPlease enter a valid number of lockers.");
                System.out.print("Enter the number of lockers: ");
                input = getIntegerInput();
            }

            // Initializes the lockers field to the provided input size.
            lockers = new Locker[input];
        } else if (option == 2) { // Runs the algorithm on the method and prints out the result.

            // Checks if the number of lockers were set or not.
            if (lockers == null) {
                System.out.println("Please enter a number of lockers prior to running the algorithm.\n");
                return; // Stops the algorithm from running on a null array.
            }

            // Sets all the lockers to closed and initializes all variables.
            for (int i = 0; i < lockers.length; i++) {
                lockers[i] = new Locker();
            }

            performAlgorithm(); // Starts the locker problem algorithm.

            // Calculates the statistics (Comparison, open/closed lockers)
            String openLockers = "";
            String closedLockers = "";

            for (int i = 1; i <= lockers.length; i++) {
                if (lockers[i-1].isOpen) {
                    openLockers += i + " ";
                } else {
                    closedLockers += i + " ";
                }
            }

            // Prints the output.
            System.out.println("Number of lockers:      " + lockers.length);
            System.out.println("Open lockers:           " + openLockers);
            System.out.println("Closed lockers:         " + closedLockers);
            System.out.println("Comparisons:            " + comparisons);
        }

        System.out.println();
    }

    /**
     * Displays the menu options.
     */
    private void displayMenu() {
        System.out.println("-----------------MAIN MENU--------------");
        System.out.println("1. Read number pf lockers (integer value)");
        System.out.println("2. Run algorithm and display output");
        System.out.println("3. Exit program");
        System.out.println();
    }

    /**
     * Runs the lockers algorithm and begins opening and closing the respective doors.
     * After every toggle, the comparison operated is incremented as that is the basic operation.
     */
    public void performAlgorithm() {
        if (lockers == null || lockers.length == 0) { // Prevents code from being run on a null array.
            return;
        }

        comparisons = 0;
        // Runs the algorithm n times through the entire list.
        for (int base = 1; base <= lockers.length; base++) {
            // Controls which lockers to flip. Goes from the base index to n, incremented by base.
            for (int toggle = base-1; toggle < lockers.length; toggle=toggle+base) {
                // Flips the locker door.
                lockers[toggle].toggle(); // Basic Operation
                comparisons++;
            }
        }
    }
}

// Made my locker list easier to follow by toggling.
class Locker {
    public boolean isOpen;

    Locker() { // Sets all new lockers to false.
        isOpen = false;
    }

    // Allowed for simpler switching.
    public void toggle() {
        isOpen = !isOpen;
    }
}